#include "graphics.h"
#include "world.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Project globals
int W_WIDTH, W_HEIGHT, W_HMID, W_VMID;
int TABLESIZE, FOV_L, D0, D90, D180, D270, CIRCLE;
float sin_t[MAXTABLESIZE], cos_t[MAXTABLESIZE], tan_t[MAXTABLESIZE];
float tan_recip_t[MAXTABLESIZE];
float Zbuffer[MAXWORLDWIDTH];
SDL_Surface *pSDL;

// Local globals
SDL_PixelFormat bmp_fmt;
unsigned int Rshift, Gshift, Bshift;
unsigned int Rmask, Gmask, Bmask;
unsigned int bpp, Bpp;
unsigned int R_off, G_off, B_off;
unsigned int Rloss, Gloss, Bloss;

// Pointers to drawing functions
// [see graphics.h for an explanation as to what's going on here]
void (*texturecol)(int x, int y, int h, float tx, int th, float d, int dc, SDL_Surface *tex);
void (*plotpixel)(int x, int y, int r, int g, int b);
void (*pixelspan)(int x, int y, int w, int r, int g, int b);
void (*sky_gradient)(int x, int h);
void (*ground_gradient)(int x, int h);
void (*pixelcol)(int x, int y, int h, int r, int g, int b);
void (*DrawBox)(int x, int y, int w, int h, int r, int g, int b);
void (*DrawBox_d)(int x, int y, int w, int h, int dr, int dg, int db);
void (*DrawObject)(int u, int v, int w, float dist, SDL_Surface *pixmap);

// Star data
int starindex[MAXTABLESIZE];
starspec stars[NUMSTARS];
int comparestars(const void *a, const void *b)
{  return ((starspec *)a)->x - ((starspec *)b)->x; }

// Gradient data
#define GRADSIZE (1 + MAXWORLDHEIGHT/2)
Uint8 skycolors[GRADSIZE][3];
Uint8 gndcolors[GRADSIZE][3];
Uint32 skypixc[GRADSIZE], gndpixc[GRADSIZE];

// Helpful inline functions
inline Uint16 MapRGB2(Uint8 r, Uint8 g, Uint8 b)
{
    return ((r >> Rloss) << Rshift) | ((g >> Gloss) << Gshift) | ((b >> Bloss) << Bshift);
}

inline Uint32 MapRGB4(Uint8 r, Uint8 g, Uint8 b)
{
    return (r << Rshift) | (g << Gshift) | (b << Bshift);
}

inline void GetRGB2(Uint16 pix, int& r, int& g, int& b)
{
    r = ((pix&Rmask) >> Rshift) << Rloss;
    g = ((pix&Gmask) >> Gshift) << Gloss;
    b = ((pix&Bmask) >> Bshift) << Bloss;
}

inline int clamp(int val)
{
    return (val < 0) ? 0 : ((val > 255) ? 255 : val);
}

inline int attenuate(int &r, int &g, int &b, int att)
{
    if(r > ATTR)
    { r -= att; if(r < ATTR) r = ATTR; }
    else
    { r += att; if(r > ATTR) r = ATTR; }

    if(g > ATTG)
    { g -= att; if(g < ATTG) g = ATTG; }
    else
    { g += att; if(g > ATTG) g = ATTG; }

    if(b > ATTB)
    { b -= att; if(b < ATTB) b = ATTB; }
    else
    { b += att; if(b > ATTB) b = ATTB; }
}


// Initializes the SDL
int init_graphics(const char *caption, int w, int h, int reqbpp, int fFullScreen)
{
    W_WIDTH = W_HEIGHT = W_VMID = 0;

    // cap w at maximum world width
    if(w > MAXWORLDWIDTH)
        w = MAXWORLDWIDTH;

    // round down w to an acceptable value
    // [ w * 90 / FOV must be an integer ]
    while( (w * 90) % FOV )
        w--;

    // now bring h up to an acceptable value
    // [ must be at least 3/4 w ]
    if(h < (3*w)/4) h = (3*w)/4;

    // Initialize SDL
    SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);
    atexit(SDL_Quit);

    if(fFullScreen)
        pSDL=SDL_SetVideoMode(w, h, reqbpp, SDL_FULLSCREEN | SDL_SWSURFACE);  // hardware surfaces are not worthwhile for my purposes
    else
        pSDL=SDL_SetVideoMode(w, h, reqbpp, SDL_SWSURFACE);

    if(!pSDL) {
        printf("init_graphics(): Couldn't initialize the window!\n");
        return 0;
    }
    if(pSDL->format->BitsPerPixel < 15) {
        printf("This program must be run in a high-color video mode (>= 15 bits per pixel).\n");
        return 0;
    }

    bpp = pSDL->format->BitsPerPixel;
    Bpp = pSDL->format->BytesPerPixel;
    Rmask = pSDL->format->Rmask; Gmask = pSDL->format->Gmask; Bmask = pSDL->format->Bmask;
    Rshift = pSDL->format->Rshift; Gshift = pSDL->format->Gshift; Bshift = pSDL->format->Bshift;
    Rloss = pSDL->format->Rloss; Gloss = pSDL->format->Gloss; Bloss = pSDL->format->Bloss;
    R_off = Rshift/8; G_off = Gshift/8; B_off = Bshift/8;

    // Choose appropriate rendering functions for the current pixel format
    switch(Bpp) // the only one of these switches left!
    {
    case 2:
        texturecol = texturecol16;
        plotpixel = plotpixel16;
        pixelspan = pixelspan16;
        sky_gradient = sky_gradient16;
        ground_gradient = ground_gradient16;
        pixelcol = pixelcol16;
        DrawBox = DrawBox16;
        DrawBox_d = DrawBox_d16;
        DrawObject = DrawObject16;
        break;
    case 3:
        texturecol = texturecol2432;
        plotpixel = plotpixel24;
        pixelspan = pixelspan24;
        sky_gradient = sky_gradient24;
        ground_gradient = ground_gradient24;
        pixelcol = pixelcol24;
        DrawBox = DrawBox24;
        DrawBox_d = DrawBox_d2432;
        DrawObject = DrawObject2432;
        break;
    case 4:
        texturecol = texturecol2432;
        plotpixel = plotpixel32;
        pixelspan = pixelspan32;
        sky_gradient = sky_gradient32;
        ground_gradient = ground_gradient32;
        pixelcol = pixelcol32;
        DrawBox = DrawBox32;
        DrawBox_d = DrawBox_d2432;
        DrawObject = DrawObject2432;
        break;
    default:
        printf("I don't know what to do with %d bytes per pixel!\n", Bpp);
        return 0;
    }

    SDL_WM_SetCaption(caption, caption);

    printf("SDL_Surface initialized: ");
    printf("window size = %dx%d, ", w, h);
    printf("bits per pixel = %d\n", pSDL->format->BitsPerPixel);

    //printf("hardware surface: %s, double-buffered: %s\n", pSDL->flags & SDL_HWSURFACE ? "yes" : "no", pSDL->flags & SDL_DOUBLEBUF ? "yes" : "no");

    W_WIDTH = w;
    W_HEIGHT = h;
    W_HMID = W_WIDTH / 2;
    W_VMID = W_HEIGHT / 2;

    // initialize TABLESIZE, FOV_L, D0, D90, D180, D270, CIRCLE;
    D0 = 0;
    D90 = W_WIDTH * 90 / FOV;
    D180 = 2 * D90;
    D270 = 3 * D90;
    CIRCLE = 4 * D90;
    TABLESIZE = CIRCLE;
    FOV_L = W_WIDTH / 2;

    double rads;
    int i;

    // initialize trig tables
    for(i=0; i<TABLESIZE; i++)
    {
        rads = (PI/2.0) * i / D90;

        sin_t[i] = (float)sin(rads);
        cos_t[i] = (float)cos(rads);
        tan_t[i] = (float)tan(rads);
        tan_recip_t[i] = 1.0f / tan_t[i];
    }

    // initialize pixel format for bitmaps
    // NOTE:  If this is changed, the definitions
    //  BMP_R, BMP_G, BMP_B, and BMP_A in graphics.h
    //  must be changed to match!
    bmp_fmt.palette = NULL;
    bmp_fmt.BitsPerPixel = 32;
    bmp_fmt.BytesPerPixel = 4;
    bmp_fmt.Rmask = 0xFF0000;
    bmp_fmt.Gmask = 0x00FF00;
    bmp_fmt.Bmask = 0x0000FF;
    bmp_fmt.Amask = 0;
    bmp_fmt.Rshift = (Uint8) 16;
    bmp_fmt.Gshift = (Uint8) 8;
    bmp_fmt.Bshift = (Uint8) 0;
    bmp_fmt.Ashift = 0;
    bmp_fmt.Rloss = (Uint8) 0;
    bmp_fmt.Gloss = (Uint8) 0;
    bmp_fmt.Bloss = (Uint8) 0;
    bmp_fmt.Aloss = (Uint8) 8;
    bmp_fmt.colorkey = 0;
    bmp_fmt.alpha = (Uint8) 0;

    // initialize sky and ground gradients
    int gradheight = 1 + W_HEIGHT/2;
    int ir, ig, ib, att;
    float r, g, b, dr, dg, db;

    r = (float)SKYR; g = (float)SKYG; b = (float)SKYB;
    dr = (ATTR - r) / gradheight; dg = (ATTG - g) / gradheight; db = (ATTB - b) / gradheight;
    for(i = 0; i < gradheight; i++)
    {
        r += dr; g += dg;  b += db;
        ir = (int)r; ig = (int)g; ib = (int)b;
        skycolors[i][0] = ir; skycolors[i][1] = ig; skycolors[i][2] = ib;
        skypixc[i] = SDL_MapRGB(pSDL->format, ir, ig, ib);
    }

    for(i = 0; i < gradheight; i++)
    {
        ir = GROUNDR; ig = GROUNDG; ib = GROUNDB;
        att = (6 * W_HEIGHT) / (ASPECT * (gradheight - i));
        attenuate(ir, ig, ib, att);
        gndcolors[i][0] = ir; gndcolors[i][1] = ig; gndcolors[i][2] = ib;
        gndpixc[i] = SDL_MapRGB(pSDL->format, ir, ig, ib);
    }

    // initialize stars

    // create random stars
    for(i = 0; i < NUMSTARS; i++) {
       stars[i].x = rand() % CIRCLE;
       stars[i].y = rand() % W_VMID;
       stars[i].i = 64 + rand() % 192;
    }

    // sort stars by increasing x-coordinate
    qsort(stars, NUMSTARS, sizeof(stars[0]), comparestars);

    // build starindex array
    int j = 0;
    for(i = 0; i < NUMSTARS; i++)
       while(j <= stars[i].x)
          starindex[j++] = i;
    for(; j < CIRCLE; j++)
       starindex[j] = 0;

    return 1;
}

// Clears the screen to black, or, more precisely, to 0,
// which is black in any pixel format I'm interested in...
void graf_clearscreen(int fRefresh)
{
    int i;
    unsigned char *p = (unsigned char *) pSDL->pixels;

    for(i=0; i < W_HEIGHT; i++)
    {
        memset(p, 0, W_WIDTH * Bpp);
        p += pSDL->pitch;
    }

    if(fRefresh)
        SDL_Flip(pSDL);
}

// This fadeout routine isn't really guaranteed to work,
// because it assumes the contents of the frame buffer
// don't change after an SDL_Flip(), which is reasonable
// since I'm using a software surface, but is by no means
// guaranteed...
void graf_fadeout(int inc)
{
    Uint8 *pix;
    int x, y, f;
    int r, g, b;
    int d = pSDL->pitch - (pSDL->w * Bpp);

    if(Bpp == 2)
    for(f = 0; f < 192/inc; f++) {
       pix = (Uint8 *)pSDL->pixels;
       SDL_LockSurface(pSDL);
       for(y = 0; y < pSDL->h; y++) {
          for(x = 0; x < pSDL->w; x++) {
             GetRGB2(*(Uint16 *)pix, r, g, b);
             r -= inc;  if(r < 0) r=0;
             g -= inc;  if(g < 0) g=0;
             b -= inc;  if(b < 0) b=0;
             *(Uint16 *)pix = MapRGB2(r, g, b);
             pix += Bpp;
          }
          pix += d;
       }
       SDL_UnlockSurface(pSDL);
       SDL_Flip(pSDL);
    }
    else
    for(f = 0; f < 192/inc; f++) {
       pix = (Uint8 *)pSDL->pixels;
       SDL_LockSurface(pSDL);
       for(y = 0; y < pSDL->h; y++) {
          for(x = 0; x < pSDL->w; x++) {
             r = pix[R_off] - inc;
             g = pix[G_off] - inc;
             b = pix[B_off] - inc;
             pix[R_off] = (r < 0) ? 0 : r;
             pix[G_off] = (g < 0) ? 0 : g;
             pix[B_off] = (b < 0) ? 0 : b;
             pix += Bpp;
          }
          pix += d;
       }
       SDL_UnlockSurface(pSDL);
       SDL_Flip(pSDL);
    }

    graf_clearscreen(true);
}

// Draws a test pattern
void ColorTest()
{
    // color test
    int i, j, k, bw=W_HEIGHT/8;
    unsigned char r, g, b;
    for(i=0; i<8; i++)
    {
        b = (i&1) ? 0xFF : 0;
        g = (i&2) ? 0xFF : 0;
        r = (i&4) ? 0xFF : 0;

        // test pixelcol
        k = i * bw;
        for(j=0; j<W_HMID; j++)
            pixelcol(j, k, bw, 2*(j * r) / W_WIDTH, 2*(j * g) / W_WIDTH, 2*(j * b) / W_WIDTH);

        // test pixelspan
        pixelspan(0, i*bw, W_HMID, 255, 255, 255);

    }

    int gw = W_WIDTH / 16, gh = W_HEIGHT / 8;
    for(i=0; i<8; i++) {
        for(j=0; j<8; j++) {
            g = ((i+j) << 4);
            r = (i << 5);
            b = (j << 5);

            // test DrawBox
            DrawBox(i*gw+W_HMID, j*gh, gw, gh, r, g, b);
        }
    }

    // test DrawLine, PlotPixel
    int wmin = W_WIDTH < W_HEIGHT ? W_WIDTH : W_HEIGHT;
    int trans_x = W_HMID, trans_y = W_VMID;

    double angle;
    double rad0=wmin/10.0, rad=wmin/2.4;
    const double pi=3.14159265358979;
    int c;

    for(angle=0; angle < 2.0 * pi; angle += pi/12.0)
    {
        c = (int)(255.0 * angle / (2*pi));
        DrawLine(trans_x + (int)(cos(angle) * rad0),
                 trans_y + (int)(sin(angle) * rad0),
                 trans_x + (int)(cos(angle) * rad),
                 trans_y + (int)(sin(angle) * rad),
                 c, c, c);
    }
}

// Draws a line using the midpoint line algorithm
void DrawLine(int x0, int y0, int x1, int y1, int r, int g, int b)
{
    // set (x0, y0) to the leftmost point
    if(x0 > x1) {
        x0 ^= x1; x1 ^= x0; x0 ^= x1;
        y0 ^= y1; y1 ^= y0; y0 ^= y1;
    }

    // Compute dx and dy
    int dx = x1 - x0;		// dx is always >= 0, since points were switched otherwise
    int dy = y1 - y0;		// dy can be anything
    int x = x0;
    int y = y0;

    plotpixel(x, y, r, g, b);	// plot the original pixel

    if(dy >= 0)	// slope is zero or positive
    {
        if(dy <= dx)	// slope is <= 1; crawl along X axis
        {
            int d = 2 * dy - dx;
            int incrE = 2 * dy;
            int incrNE = 2 * (dy - dx);

            for(; x < x1; x++) {
                if(d <= 0) {
                    d += incrE;
                }
                else {
                    d += incrNE;
                    y++;
                }
                plotpixel(x, y, r, g, b);
            }
        }
        else	// slope is > 1; crawl along Y axis
        {
            int d = 2 * dx - dy;
            int incrN = 2 * dx;
            int incrNE = 2 * (dx - dy);

            for(; y < y1; y++) {
                if(d <= 0) {
                    d += incrN;
                }
                else {
                    d += incrNE;
                    x++;
                }
                plotpixel(x, y, r, g, b);
            }
        }
    }
    else	// slope is negative
    {
        dy = -dy;		// work with the magnitude of dy

        if(dy <= dx)	// crawl along X axis
        {
            int d = 2 * dy - dx;
            int incrE = 2 * dy;
            int incrSE = 2 * (dy - dx);

            for(; x < x1; x++) {
                if(d <= 0) {
                    d += incrE;
                }
                else {
                    d += incrSE;
                    y--;
                }
                plotpixel(x, y, r, g, b);
            }
        }
        else		// crawl along Y axis
        {
            int d = 2 * dx - dy;
            int incrS = 2 * dx;
            int incrSE = 2 * (dx - dy);

            for(; y > y1; y--) {
                if(d <= 0) {
                    d += incrS;
                }
                else {
                    d += incrSE;
                    x++;
                }
                plotpixel(x, y, r, g, b);
            }
        }  // if [slope magnitude]
    }	// if [slope sign]
}	// DrawLine


// Encodes a bitmap in a 32-bit format, setting the alpha channel
// of all pixels matching the upper-left one to transparent.
SDL_Surface *Encode32(SDL_Surface *pixmap, int alpha)
{
    // bmp_fmt is a 32-bpp format; we'll set alpha here if desired
    SDL_Surface *pNew = SDL_ConvertSurface(pixmap, &bmp_fmt, SDL_SWSURFACE);
    SDL_FreeSurface(pixmap);

    if(alpha) {
        Uint8 *pTex = (Uint8 *)pNew->pixels;
        Uint32 trans = *(Uint32 *)pTex, texel;
        int inc = pNew->pitch - 4 * pNew->w;
        int i, j;

        for(i=0; i<pNew->h; i++)
        {
            for(j=0; j<pNew->w; j++)
            {
                texel = *(Uint32 *)pTex;
                if(texel==trans)
                    pTex[BMP_A] = SDL_ALPHA_TRANSPARENT;
                else
                    pTex[BMP_A] = SDL_ALPHA_OPAQUE;
                pTex += 4;
            }
            pTex += inc;
        }
    }

    return pNew;
}


// texturecol: the basic texture mapping function.
//   (x, y) = pixel coordinates of the top of the column
//    h     = column height in pixels [should be capped at W_HEIGHT]
//    tx    = horizontal offset in texture, in the range [0..1]
//    th    = wall height in pixels [can be > W_HEIGHT]
//    d     = perpendicular distance to wall
//    dc    = pseudo-diffuse damping color
//    tex   = pointer to 32-bit SDL_Surface containing texture

void texturecol16(int x, int y, int h, float tx, int th, float d, int dc, SDL_Surface *tex)
{
    int att = (int)(d * ATTF);
    int col = (int)(tx * tex->w);
    int v, rem = 0;
    v = (th==h) ? 0 : ((th - h) * tex->h) / (th << 1);
    int tex_h = tex->h - (v << 1);
    int r, g, b, i;

    Uint8 *pTex = (Uint8 *)tex->pixels + tex->pitch * v + 4 * col;
    Uint8 *pixloc=((Uint8 *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;

    int num = tex_h-1, den = h-1;
    int inc = den;

    for(i = 0; i < h; i++)
    {
        // Get a texel
        r = pTex[BMP_R] - dc; if(r < 0) r = 0;
        g = pTex[BMP_G] - dc; if(g < 0) g = 0;
        b = pTex[BMP_B] - dc; if(b < 0) b = 0;

        // Perform atmospheric attenuation
        attenuate(r, g, b, att);

        // Write a pixel and advance the screen pointer
        *(Uint16 *)pixloc = MapRGB2(r, g, b);
        pixloc += pSDL->pitch;

        // Advance the texture pointer if needed
        inc += num;
        while(inc > den)
        {
            pTex += tex->pitch;
            inc -= den;
        }
    }
}

void texturecol2432(int x, int y, int h, float tx, int th, float d, int dc, SDL_Surface *tex)
{
    int att = (int)(d * ATTF);
    int col = (int)(tx * tex->w);
    int v, rem = 0;
    v = (th==h) ? 0 : ((th - h) * tex->h) / (th << 1);
    int tex_h = tex->h - (v << 1);
    int r, g, b, i;

    Uint8 *pTex = (Uint8 *)tex->pixels + tex->pitch * v + 4 * col;
    Uint8 *pixloc=((Uint8 *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;

    int num = tex_h-1, den = h-1;
    int inc = den;

    for(i = 0; i < h; i++)
    {
        // Get a texel
        r = pTex[BMP_R] - dc; if(r < 0) r = 0;
        g = pTex[BMP_G] - dc; if(g < 0) g = 0;
        b = pTex[BMP_B] - dc; if(b < 0) b = 0;

        // Perform atmospheric attenuation
        attenuate(r, g, b, att);

        // Write a pixel and advance the screen pointer
        pixloc[R_off] = r; pixloc[G_off] = g; pixloc[B_off] = b;
        pixloc += pSDL->pitch;

        // Advance the texture pointer if needed
        inc += num;
        while(inc > den)
        {
            pTex += tex->pitch;
            inc -= den;
        }
    }
}

// plotpixel: plots a single pixel at (x, y) having color (r, g, b)

void plotpixel16(int x, int y, int r, int g, int b)
{
    // do simple bounds checking and avoid nasty core dumps
    if((Uint32)x >= (Uint32)W_WIDTH || (Uint32)y >= (Uint32)W_HEIGHT)
        return;

    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    *(Uint16 *)pixloc = MapRGB2(r, g, b);
}

void plotpixel24(int x, int y, int r, int g, int b)
{
    // do simple bounds checking and avoid nasty core dumps
    if((Uint32)x >= (Uint32)W_WIDTH || (Uint32)y >= (Uint32)W_HEIGHT)
        return;

    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    pixloc[R_off] = r; pixloc[G_off] = g; pixloc[B_off] = b;
}

void plotpixel32(int x, int y, int r, int g, int b)
{
    // do simple bounds checking and avoid nasty core dumps
    if((Uint32)x >= (Uint32)W_WIDTH || (Uint32)y >= (Uint32)W_HEIGHT)
        return;

    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    *(Uint32 *)pixloc = MapRGB4(r, g, b);
}

// pixelspan: plots a row of pixels more efficiently than multiple calls to plotpixel
// DOES NOT DO BOUNDS CHECKING!

void pixelspan16(int x, int y, int w, int r, int g, int b)
{
    Uint16 *pixloc=(Uint16 *)((unsigned char *)pSDL->pixels + pSDL->pitch * y + Bpp * x);
    Uint16 pixel=MapRGB2(r, g, b);

    for(int i=0; i<w; i++)
        *(pixloc++) = pixel;
}

void pixelspan24(int x, int y, int w, int r, int g, int b)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    for(int i=0; i<w; i++) {
        pixloc[R_off] = r;
        pixloc[G_off] = g;
        pixloc[B_off] = b;
        pixloc += Bpp;
    }
}

void pixelspan32(int x, int y, int w, int r, int g, int b)
{
    Uint32 *pixloc=(Uint32 *)((unsigned char *)pSDL->pixels + pSDL->pitch * y + Bpp * x);
    Uint32 pixel=MapRGB4(r, g, b);

    for(int i=0; i<w; i++)
        *(pixloc++) = pixel;
}

// sky_gradient:  Appropriately enough, this draws the gradient in the sky.
//   x = column, h = height
//   Colors are initialized in init_graphics.

void sky_gradient16(int x, int h)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + Bpp * x;

    for(int col = 0; col < h; col++) {
        *(Uint16 *)pixloc = (Uint16)skypixc[col];
        pixloc += pSDL->pitch;
    }
}

void sky_gradient24(int x, int h)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + Bpp * x;

    for(int col = 0; col < h; col++) {
        pixloc[R_off] = skycolors[col][0]; pixloc[G_off] = skycolors[col][1]; pixloc[B_off] = skycolors[col][2];
        pixloc += pSDL->pitch;
    }
}

void sky_gradient32(int x, int h)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + Bpp * x;

    for(int col = 0; col < h; col++) {
        *(Uint32 *)pixloc = skypixc[col];
        pixloc += pSDL->pitch;
    }
}

// ground_gradient: Self-explanatory.
//   x = column; h = height
//   Colors are initialized in init_graphics.

void ground_gradient16(int x, int h)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + Bpp * x + pSDL->pitch * (W_HEIGHT-1);
    int col;

    for(col = 0; col < h; col++) {
        *(Uint16 *)pixloc = (Uint16)gndpixc[col];
        pixloc -= pSDL->pitch;
    }
}

void ground_gradient24(int x, int h)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + Bpp * x + pSDL->pitch * (W_HEIGHT-1);
    int col;

    for(col = 0; col < h; col++) {
        pixloc[R_off] = gndcolors[col][0]; pixloc[G_off] = gndcolors[col][1]; pixloc[B_off] = gndcolors[col][2];
        pixloc -= pSDL->pitch;
    }
}

void ground_gradient32(int x, int h)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + Bpp * x + pSDL->pitch * (W_HEIGHT-1);
    int col;

    for(col = 0; col < h; col++) {
        *(Uint32 *)pixloc = gndpixc[col];
        pixloc -= pSDL->pitch;
    }
}

// pixelcol: Draws a column of pixels h tall, starting at (x, y), more efficiently than multiple calls to PlotPixel
// DOES NOT DO BOUNDS CHECKING!
void pixelcol16(int x, int y, int h, int r, int g, int b)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    Uint16 pixel;
    int col;

    pixel = MapRGB2(r, g, b);
    for(col = 0; col < h; col++) {
        *(Uint16 *)pixloc = pixel;
        pixloc += pSDL->pitch;
    }
}

void pixelcol24(int x, int y, int h, int r, int g, int b)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;

    for(int col = 0; col < h; col++) {
        pixloc[R_off] = r; pixloc[G_off] = g; pixloc[B_off] = b;
        pixloc += pSDL->pitch;
    }
}

void pixelcol32(int x, int y, int h, int r, int g, int b)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    Uint32 pixel;
    int col;

    pixel = MapRGB4(r, g, b);
    for(col = 0; col < h; col++) {
        *(Uint32 *)pixloc = pixel;
        pixloc += pSDL->pitch;
    }
}


// DrawBox: Draws a box more efficiently than multiple calls to pixelspan or (heaven forbid) plotpixel
// DOES NOT DO BOUNDS CHECKING!
void DrawBox16(int x, int y, int w, int h, int r, int g, int b)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    int i, row;
    Uint16 pixel=MapRGB2(r, g, b);

    for(row = 0; row < h; row++)
    {
        for(i=0; i<w; i++)
            *((Uint16 *)pixloc + i) = pixel;
        pixloc += pSDL->pitch;
    }
}

void DrawBox24(int x, int y, int w, int h, int r, int g, int b)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    int i, row, delta = pSDL->pitch - Bpp * w;

    for(row = 0; row < h; row++)
    {
        for(i=0; i<w; i++) {
            pixloc[R_off] = r;
            pixloc[G_off] = g;
            pixloc[B_off] = b;
            pixloc += Bpp;
        }
        pixloc += delta;
    }
}

void DrawBox32(int x, int y, int w, int h, int r, int g, int b)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    int i, row;
    Uint32 pixel=MapRGB4(r, g, b);

    for(row = 0; row < h; row++)
    {
        for(i=0; i<w; i++)
            *((Uint32 *)pixloc + i) = pixel;
        pixloc += pSDL->pitch;
    }
}

// DrawBox_d: Draws a translucent box
void DrawBox_d16(int x, int y, int w, int h, int dr, int dg, int db)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    int i, row;
    unsigned int pixel;

    int r, g, b;
    for(row = 0; row < h; row++) {
        for(i=0; i<w; i++) {
            pixel = (unsigned long) *((unsigned short *)pixloc + i);
            GetRGB2(pixel, r, g, b);
            r = clamp(r + dr); g = clamp(g + dg); b = clamp(b + db);
            pixel=MapRGB2(r, g, b);
            *((unsigned short *)pixloc + i) = (unsigned short)pixel;
        }
        pixloc += pSDL->pitch;
    }
}

void DrawBox_d2432(int x, int y, int w, int h, int dr, int dg, int db)
{
    unsigned char *pixloc=((unsigned char *)pSDL->pixels) + pSDL->pitch * y + Bpp * x;
    int i, row, delta;

    delta = pSDL->pitch - Bpp * w;
    for(row = 0; row < h; row++)
    {
        for(i=0; i<w; i++) {
            pixloc[R_off] = clamp(pixloc[R_off]+dr);
            pixloc[G_off] = clamp(pixloc[G_off]+dg);
            pixloc[B_off] = clamp(pixloc[B_off]+db);
            pixloc += Bpp;
        }
        pixloc += delta;
    }
}

// DrawObject: Draws an object
//   (u, v) = pixel coordinates (objects are clipped horizontally but NOT vertically)
//   w = bitmap width, height [w must be < min(W_HEIGHT, W_WIDTH)]
//   dist = perpendicular distance to object (compared against Z-buffer values at each column)
//   pixmap = pointer to 32-bit SDL_Surface containing the object's pixmap

void DrawObject16(int u, int v, int w, float dist, SDL_Surface *pixmap)
{
    Uint8 *pTex;
    int h = w, tw = pixmap->w, th = pixmap->h;
    int att = (int)(dist * ATTF);
    int r, g, b;

    // clip the left edge if needed
    if(u < 0) {
        int dt = (u * tw) / w;
        w += u;
        tw += dt;
        pTex = (Uint8 *)pixmap->pixels - 4 * dt;
        u = 0;
    }
    else pTex = (Uint8 *)pixmap->pixels;
    Uint8 *pTexCol;
    Uint8 *pPix = (Uint8 *)pSDL->pixels + pSDL->pitch * v + Bpp * u;
    Uint8 *pPixCol;

    // Now we have a (tw x th) pixmap to be mapped to a (w x h) screen window
    int j;
    int n0 = tw-1, d0 = w-1, i0 = d0;
    int num, den, inc;

    // clip right edge if needed
    int xf = u + w - 1;  if(xf >= W_WIDTH) xf = W_WIDTH;

    for(; u < xf; u++)
    {
        // Check for object visibility
        if(dist < Zbuffer[u])
        {
            pPixCol = pPix;
            pTexCol = pTex;

            // Write a column
            num = th-1; den = h-1; inc = den;
            for(j = 0; j < h; j++)
            {
                // Write a pixel...
                if(pTexCol[BMP_A] != SDL_ALPHA_TRANSPARENT)
                {
                    r = pTexCol[BMP_R];
                    g = pTexCol[BMP_G];
                    b = pTexCol[BMP_B];

                    attenuate(r, g, b, att);
                    *(Uint16 *)pPixCol = MapRGB2(r, g, b);
                }

                // ...and advance the screen pointer
                pPixCol += pSDL->pitch;

                // Advance the texture pointer if needed
                inc += num;
                while(inc > den)
                {
                    pTexCol += pixmap->pitch;
                    inc -= den;
                }
            }
        }

        // Handle single-pixel width correctly
        if(d0==0) break;

        // Advance the screen pointer
        pPix += Bpp;

        // Advance the texture pointer if needed
        i0 += n0;
        while(i0 > d0)
        {
            pTex += 4;
            i0 -= d0;
        }
    }
}

void DrawObject2432(int u, int v, int w, float dist, SDL_Surface *pixmap)
{
    Uint8 *pTex;
    int h = w, tw = pixmap->w, th = pixmap->h;
    int att = (int)(dist * ATTF);

    // clip the left edge if needed
    if(u < 0) {
        int dt = (u * tw) / w;
        w += u;
        tw += dt;
        pTex = (Uint8 *)pixmap->pixels - 4 * dt;
        u = 0;
    }
    else pTex = (Uint8 *)pixmap->pixels;
    Uint8 *pTexCol;
    Uint8 *pPix = (Uint8 *)pSDL->pixels + pSDL->pitch * v + Bpp * u;
    Uint8 *pPixCol;

    // Now we have a (tw x th) pixmap to be mapped to a (w x h) screen window
    int j;
    int n0 = tw-1, d0 = w-1, i0 = d0;
    int num, den, inc;
    int r, g, b;

    // clip right edge if needed
    int xf = u + w - 1;  if(xf >= W_WIDTH) xf = W_WIDTH;

    for(; u < xf; u++)
    {
        // Check for object visibility
        if(dist < Zbuffer[u])
        {
            pPixCol = pPix;
            pTexCol = pTex;

            // Write a column
            num = th-1; den = h-1; inc = den;
            for(j = 0; j < h; j++)
            {
                // Write a pixel...
                if(pTexCol[BMP_A] != SDL_ALPHA_TRANSPARENT)
                {
                    r = pTexCol[BMP_R];
                    g = pTexCol[BMP_G];
                    b = pTexCol[BMP_B];

                    attenuate(r, g, b, att);

                    pPixCol[R_off] = r;
                    pPixCol[G_off] = g;
                    pPixCol[B_off] = b;
                }

                // ...and advance the screen pointer
                pPixCol += pSDL->pitch;

                // Advance the texture pointer if needed
                inc += num;
                while(inc > den)
                {
                    pTexCol += pixmap->pitch;
                    inc -= den;
                }
            }
        }

        // Handle single-pixel width correctly
        if(d0==0) break;

        // Advance the screen pointer
        pPix += Bpp;

        // Advance the texture pointer if needed
        i0 += n0;
        while(i0 > d0)
        {
            pTex += 4;
            i0 -= d0;
        }
    }
}

