#ifndef GRAPHICS_H
#define GRAPHICS_H

#include <SDL/SDL.h>
#include <math.h>

// Allow compiling with older SDL versions
#ifndef SDL_ALPHA_TRANSPARENT
#define SDL_ALPHA_TRANSPARENT ((Uint8)0)
#endif
#ifndef SDL_ALPHA_OPAQUE
#define SDL_ALPHA_OPAQUE ((Uint8)255)
#endif

// Constant view parameters
#define FOV             60                         // field of vision, in degrees
#define MAXWORLDWIDTH   1280                       // maximum window size
#define MAXWORLDHEIGHT  ((MAXWORLDWIDTH * 3)/4)
#define MAXTABLESIZE    ((360*MAXWORLDWIDTH)/FOV)  // maximum size of trig tables
#define NUMSTARS        300

// Byte offsets in loaded bitmaps
#define BMP_A 3
#define BMP_R 2
#define BMP_G 1
#define BMP_B 0


// Sky, ground, and attenuation colors
// Night profile
#define SKYR  0
#define SKYG  64
#define SKYB  96
#define GROUNDR  48
#define GROUNDG  32
#define GROUNDB  12
#define ATTR 0
#define ATTG 0
#define ATTB 32
#define ATTF 16	// attenuation factor; higher values mean steeper attenuation

// Global variables

// Viewport dimensions
extern int W_WIDTH, W_HEIGHT, W_HMID, W_VMID;

// Trig table dimensions
extern int TABLESIZE, FOV_L, D0, D90, D180, D270, CIRCLE;

// Trig tables
extern float sin_t[MAXTABLESIZE], cos_t[MAXTABLESIZE], tan_t[MAXTABLESIZE];
extern float tan_recip_t[MAXTABLESIZE];

// The Z-buffer
extern float Zbuffer[MAXWORLDWIDTH];

// Generic function prototypes
int init_graphics(const char *caption, int reqW, int reqH, int reqBPP, int fFullScreen);
void graf_clearscreen(int fRefresh);
void graf_fadeout(int inc);
void ColorTest();
void DrawLine(int x0, int y0, int x1, int y1, int r, int g, int b);
SDL_Surface *Encode32(SDL_Surface *pixmap, int alpha = false);

// The following functions differ based on the pixel format of the display.
// Thus there are two or three versions of each function.  The pointer
// which is declared above each function group is set by init_graphics to
// point at whichever function is appropriate for the given bit depth.

// Note that the different versions are usually nearly identical; the reason
// I use separate functions rather than one generalized one is that I'd rather
// not test which version is appropriate more than once, let alone
// hundreds of times per frame.

extern void (*texturecol)(int x, int y, int h, float tx, int th, float d, int dc, SDL_Surface *tex);
void texturecol16(int x, int y, int h, float tx, int th, float d, int dc, SDL_Surface *tex);
void texturecol2432(int x, int y, int h, float tx, int th, float d, int dc, SDL_Surface *tex);

extern void (*plotpixel)(int x, int y, int r, int g, int b);
void plotpixel16(int x, int y, int r, int g, int b);
void plotpixel24(int x, int y, int r, int g, int b);
void plotpixel32(int x, int y, int r, int g, int b);

extern void (*pixelspan)(int x, int y, int w, int r, int g, int b);
void pixelspan16(int x, int y, int w, int r, int g, int b);
void pixelspan24(int x, int y, int w, int r, int g, int b);
void pixelspan32(int x, int y, int w, int r, int g, int b);

extern void (*sky_gradient)(int x, int h);
void sky_gradient16(int x, int h);
void sky_gradient24(int x, int h);
void sky_gradient32(int x, int h);

extern void (*ground_gradient)(int x, int h);
void ground_gradient16(int x, int h);
void ground_gradient24(int x, int h);
void ground_gradient32(int x, int h);

extern void (*pixelcol)(int x, int y, int h, int r, int g, int b);
void pixelcol16(int x, int y, int h, int r, int g, int b);
void pixelcol24(int x, int y, int h, int r, int g, int b);
void pixelcol32(int x, int y, int h, int r, int g, int b);

extern void (*DrawBox)(int x, int y, int w, int h, int r, int g, int b);
void DrawBox16(int x, int y, int w, int h, int r, int g, int b);
void DrawBox24(int x, int y, int w, int h, int r, int g, int b);
void DrawBox32(int x, int y, int w, int h, int r, int g, int b);

extern void (*DrawBox_d)(int x, int y, int w, int h, int dr, int dg, int db);
void DrawBox_d16(int x, int y, int w, int h, int dr, int dg, int db);
void DrawBox_d2432(int x, int y, int w, int h, int dr, int dg, int db);

extern void (*DrawObject)(int u, int v, int w, float dist, SDL_Surface *pixmap);
void DrawObject16(int u, int v, int w, float dist, SDL_Surface *pixmap);
void DrawObject2432(int u, int v, int w, float dist, SDL_Surface *pixmap);


// Star definition structures
struct starspec {
   int x;
   int y;
   int i;
};
extern int starindex[MAXTABLESIZE];
extern starspec stars[NUMSTARS];

// Star drawing function
// (it's inlined because I didn't want to incur W_WIDTH extra function calls per frame for something so simple)
inline void DrawStars(int vangle, int col)
{
    int i = starindex[vangle];
    while(stars[i].x == vangle)
    {
        plotpixel(col, stars[i].y, stars[i].i, stars[i].i, stars[i].i);
        if(++i >= NUMSTARS) i=0;
    }
}

#endif //GRAPHICS_H

