#include <string.h>
#include "graphics.h"
#include "world.h"
#include "SDLFont.h"
#include "state.h"

const char font_file[] = "font9_14.bmp";

#define NUMSKILLS 4
const char skills[NUMSKILLS][30] = { "Pathetic", "Novice", "Intermediate", "Advanced" };

static char appversion[] = "Amazing v1.0 by Jeremy Stanley";
static char apptitle[] = "Amazing";

// default viewport dimensions
#define DEFAULT_WIDTH   640
#define DEFAULT_HEIGHT  480

// Angular and linear player movement speeds
#define DEGREES_PER_SECOND 120
#define BLOCKS_PER_SECOND  2.00f

// Minimum distance between the player and a wall
#define WALLD 0.15f

// Global variables
extern SDL_Surface *pSDL;	// declared in graphics.cpp
options opt;
gamestate state;
int frames = 0;
char sFPS[100] = "";
char sLine[100];
int map_bw;

// Function prototypes
int MenuLoop(World &W, SDLFont &F);
int GameLoop(World &W, SDLFont &F);
void DrawMenuText(SDLFont &F);
int HighScores(SDLFont &F);

// Timer callback function [used by the game clock as well as the frame rate counter]
Uint32 TimerCallback(Uint32 interval)
{
    if(state.fps) {
        sprintf(sFPS, "%3d fps", frames);
    }
    else sFPS[0] = '\0';

    frames = 0;
    state.time++;

    return interval;
}

// Display usage information
void usage() {
    printf("Options:\n");
    printf("  -w width    Specify window width in pixels\n");
    printf("  -h height   Specify window height in pixels (default is 3/4 width)\n");
	printf("  -b bpp      Request number of bits per pixel (default uses current display mode)\n");
    printf("  -f          Run in a full screen\n");
}

// Faster than a call to fabs()
inline float abs(float x) {
   return (x < 0) ? -x : x;
}

// The program's entry point
int main(int argc, char **argv)
{
    int w = 0, h = 0, fFullScreen = 0, i, maxarg = argc - 1, b = 0;
	printf("%s\n", appversion);

    // Process command-line arguments
    for(i=1; i<argc; i++) {
        if((0==strcmp(argv[i], "-w")) && maxarg > i)
            w=atoi(argv[++i]);
        else if((0==strcmp(argv[i], "-h")) && maxarg > i)
            h=atoi(argv[++i]);
		else if((0==strcmp(argv[i], "-b")) && maxarg > i)
			b=atoi(argv[++i]);
        else if(0==strcmp(argv[i], "-f"))
            fFullScreen = 1;
        else {
            usage();
            return 0;
        }
    }

    if(w==0) {
        w = DEFAULT_WIDTH;
        h = DEFAULT_HEIGHT;
    }
    else if(h==0) {
        h = (3*w) / 4;
    }

    if(!init_graphics(apptitle, w, h, b, fFullScreen))
        return 1;

    // Initialize a font
    SDLFont F(pSDL);
    if(!F.LoadFont(font_file)) {
        printf("Couldn't load font file %s\n", font_file);
        return 1;
    }

    World W;
    opt.skill = 0;
    opt.height = 15;
    opt.width = 15;
    map_bw = W_WIDTH / 64;

    // Start FPS timer
    SDL_SetTimer(1000, TimerCallback);

    do {
	W.CreateTitle();

        if(!MenuLoop(W, F))
            return 0;

        W.Resize(opt.width, opt.height);
        W.CreateMaze(1);
	F.SetColor(255, 0, 0);

	if(!GameLoop(W, F))
	   return 0;

        graf_fadeout(10);
    } while(HighScores(F));

    return 0;
}

// displays the main menu, and also calibrates speed
// options set are stored in the opt structure
int MenuLoop(World &W, SDLFont &F)
{
    int fc = 0;
    Uint32 start = SDL_GetTicks();

    SDL_Event se;
    int done = 0;

    while(!done || fc < 6) // require at least 6 frames to be drawn, to calibrate speed
    {
        if(SDL_PollEvent(&se))
        {
            if(se.type==SDL_KEYDOWN)
            {
                switch(se.key.keysym.sym)
                {
                case SDLK_ESCAPE:
                    return 0;

                case SDLK_RETURN:
                case SDLK_KP_ENTER:
                    done = 1;
                    break;

                case SDLK_KP_PLUS:
                case SDLK_PLUS:
                case SDLK_EQUALS:
                    if(opt.width < 45) {
                        opt.width += 2;
                        opt.height += 2;
                    }
                    break;

                case SDLK_KP_MINUS:
                case SDLK_MINUS:
                    if(opt.width > 7) {
                        opt.width -= 2;
                        opt.height -= 2;
                    }
                    break;

                case SDLK_s:
                    opt.skill = (opt.skill + 1) % NUMSKILLS;
                    break;

                }
            }
        }

        SDL_LockSurface(pSDL);
        W.Render3D(state.x, state.y, state.angle);
        W.RenderObjects();
        W.Quick2DTrans(10, 10, map_bw);
        DrawMenuText(F);
        fc++;
        SDL_UnlockSurface(pSDL);
        SDL_Flip(pSDL);
    }

    // get elapsed time and set ma and mv appropriately
    Uint32 elapsed = SDL_GetTicks() - start;
    float fps = 1000.0f * fc / elapsed;
    opt.mv = BLOCKS_PER_SECOND / fps;
    opt.ma = (int)(DTOU(DEGREES_PER_SECOND) / fps + 0.5f);

    printf("Speed calibration results: %.2f fps\n", fps);

    return 1;
}

void DrawMenuText(SDLFont &F)
{
    F.SetColor(0, 128, 192);
    F.CenterText(1, 16, "AMAZING");
    F.SetColor(255, 255, 255);
    F.CenterText(0, 15, "AMAZING");

    F.SetColor(0, 192, 255);
    F.CenterText(0, 45, "by Jeremy Stanley");

    F.SetColor(255, 192, 0);
    sprintf(sLine, "  [+/-]  Maze size: %d", opt.width);
    F.WriteText(10, W_HEIGHT - 65, sLine);
    sprintf(sLine, "    [S]  Skill: %s", skills[opt.skill]);
    F.WriteText(10, W_HEIGHT - 45, sLine);
    F.WriteText(10, W_HEIGHT - 25, "[ENTER]  Begin");
}

int GameLoop(World &W, SDLFont &F)
{
    // initialize local variables
    float dx, dy, bx, by;
    int ix, iy;
    int sideangle;
    int fpsx = W_WIDTH - 70;
    int fpsy = W_HEIGHT - 20;
    int ncolx = W_WIDTH - 120;
    int ncoly = 30;
    int timex = W_WIDTH - 120;
    int timey = 10;
    byte c;

    SDL_Event se;
    do {
        // Event handling
        if(SDL_PollEvent(&se))
        {
            if(se.type==SDL_KEYUP)
            {
                switch(se.key.keysym.sym)
                {
                case SDLK_ESCAPE:
                    return 0;

                case SDLK_LEFT:
                case SDLK_RIGHT:
                case SDLK_KP4:
                case SDLK_KP6:
                    state.w = 0;
                    break;

                case SDLK_KP8:
                case SDLK_KP2:
                case SDLK_UP:
                case SDLK_DOWN:
                case SDLK_a:
                case SDLK_z:
                    state.v = 0.0f;
                    break;

                case SDLK_KP1:
                case SDLK_KP3:
                    state.vs = 0.0f;

                case SDLK_LSHIFT:
                case SDLK_s:
                    state.run = 1.0f;
                    break;

                case SDLK_m:
                    state.map = !state.map;
                    break;

                case SDLK_f:
                    state.fps = !state.fps;
                    break;

                }
            }
            else if(se.type==SDL_KEYDOWN)
            {
                switch(se.key.keysym.sym)
                {
                case SDLK_LEFT:
                case SDLK_KP4:
                    state.w = opt.ma;
                    break;

                case SDLK_RIGHT:
                case SDLK_KP6:
                    state.w = -opt.ma;
                    break;

                case SDLK_LSHIFT:
                case SDLK_s:
                    state.run = 2.0f;
                    break;

                case SDLK_KP8:
                case SDLK_UP:
                case SDLK_a:
                    state.v = opt.mv;
                    break;

                case SDLK_KP2:
                case SDLK_DOWN:
                case SDLK_z:
                    state.v = -opt.mv;
                    break;

                case SDLK_KP1:
                    state.vs = -opt.mv;
                    break;

                case SDLK_KP3:
                    state.vs = opt.mv;
                    break;


                }
            }
        }

        // Angle updating
        state.angle += state.w;
        if(state.angle < D0) state.angle += CIRCLE;
        else if(state.angle >= CIRCLE) state.angle -= CIRCLE;

        // Position updating

        // Forward component
        dx =  state.v * state.run * cos_t[state.angle];
        dy = -state.v * state.run * sin_t[state.angle];

        // Sideways component
        sideangle = state.angle - D90;
        if(sideangle < 0) sideangle += CIRCLE;
        dx +=  state.vs * cos_t[sideangle];
        dy += -state.vs * sin_t[sideangle];

        // Collision detection
        ix = (int)state.x;
        iy = (int)state.y;

        state.x += dx;
        state.y += dy;

        if(dx > 0) {
           if(0 != (c=W.get(ix+1, iy))) {
              bx = (1.0f-WALLD) + ix;
              if(state.x > bx) {
                  if(c==3 && abs(0.5f - FPART(state.y)) < .25) return 1;
                  else state.x = bx;
              }
           }
        }
        else
        {
            if(0 != (c=W.get(ix-1, iy))) {
              bx = WALLD + ix;
              if(state.x < bx)  {
                  if(c==3 && abs(0.5f - FPART(state.y)) < .25) return 1;
                  else state.x = bx;
              }
            }
        }

        if(dy > 0) {
           if(0 != (c=W.get(ix, iy+1))) {
              by = (1.0f-WALLD) + iy;
              if(state.y > by) {
                  if(c==3 && abs(0.5f - FPART(state.x)) < .25) return 1;
                  else state.y = by;
              }
           }
        }
        else
        {
            if(0 != (c=W.get(ix, iy-1))) {
              by = WALLD + iy;
              if(state.y < by) {
                  if(c==3 && abs(0.5f - FPART(state.x)) < .25) return 1;
                  else state.y = by;
              }
            }
        }

        // Last chance collision detection--if we're inside a wall, get out!
        if(W.get((int)state.x, (int)state.y))
        {
          state.x -= dx;
          state.y -= dy;
        }

        // Screen rendering
        SDL_LockSurface(pSDL);
        // Render the world
        W.Render3D(state.x, state.y, state.angle);
        // Render objects
        W.RenderObjects();

	// Draw the map, if applicable
	if(state.map) W.Quick2DTrans(10, 10, map_bw);

	// Draw the frame rate counter, if applicable
        if(state.fps) {
            F.WriteText(fpsx, fpsy, sFPS);
        }
        frames++;

	// Draw game statistics
	sprintf(sLine, "%2d / %2d", state.collected, opt.collectables);
        F.WriteText(ncolx, ncoly, sLine);

        sprintf(sLine, "Time:%4d", state.time);
        F.WriteText(timex, timey, sLine);

        SDL_UnlockSurface(pSDL);
	SDL_Flip(pSDL);

    } while(se.type != SDL_QUIT);

    return 0;
}

// Considering the user's freedom in selecting maze size and skill level,
// I haven't come up with a high score system yet
int HighScores(SDLFont &F)
{
    char msg[100];
    strcpy(msg, "CONGRATULATIONS!  YOU'RE FREE!");

    F.SetColor(0, 0, 128);
    F.CenterText(1, 45, msg);
    F.SetColor(0, 128, 255);
    F.CenterText(0, 44, msg);
    F.SetColor(128, 255, 255);
    F.CenterText(-1, 43, msg);

    sprintf(msg, "Maze size: %d x %d", opt.width, opt.height);
    F.SetColor(255, 255, 192);
    F.CenterText(0, 100, msg);

    sprintf(msg, "Your time: %d seconds", state.time);
    F.CenterText(0, 140, msg);

	F.SetColor(255, 192, 192);
	F.CenterText(0, W_HEIGHT - 24, "Another game? (Y/N)");


    SDL_Flip(pSDL);

    int done = 0;
    SDL_Event se;

    while(!done)
    {
        if(SDL_PollEvent(&se))
        {
            if(se.type==SDL_KEYDOWN)
            {
                switch(se.key.keysym.sym)
                {
                case SDLK_ESCAPE:
				case SDLK_n:
                    return 0;
				case SDLK_RETURN:
				case SDLK_KP_ENTER:
                case SDLK_y:
					return 1;
				default:
                        ;
                }
            }
        }
    }

    return 0;
}
