#include "world.h"
#include "graphics.h"
#include "state.h"
#include <string.h>

extern SDL_Surface *pSDL;

#define LINEBUF 200
char linebuf[LINEBUF];

inline int ABS(int x)
{
    return (x < 0) ? -x : x;
}

inline float ABS(float x)
{
    return (x < 0) ? -x : x;
}

const char tex_names[NUMTEX][20] = { "texture1.bmp", "texture2.bmp", "texture3.bmp"};

inline int clamp(int val)
{
    return (val < 0) ? 0 : ((val > 255) ? 255 : val);
}

inline int clamp192(int val)
{
    return (val < 0) ? 0 : ((val > 192) ? 192 : val);
}

World::World()
{
    width = 0;
    height = 0;
    world = NULL;

    srand((unsigned)time(NULL));
    int i;

    SDL_Surface *tmp;

    // load textures
    for(i=0; i<NUMTEX; i++) {
        tmp = SDL_LoadBMP(tex_names[i]);
        if(tmp==NULL) {
            printf("Couldn't load texture '%s'\n", tex_names[i]);
            exit(1);
        }
        else {
            tex[i] = Encode32(tmp);
            printf("Loaded texture '%s'\n", tex_names[i]);
        }
    }

	// load object definitions
	FILE *objnames=fopen(OBJFILE, "r");
	if(objnames==NULL) {
		printf("Couldn't open object bitmap list '%s'!\n", OBJFILE);
		exit(1);
	}
	else printf("Loaded object bitmap list '%s':\n", OBJFILE);
	
	char *name;
	opt.numobjtypes = 0;
	while(!feof(objnames)) {
		if(!fgets(linebuf, LINEBUF, objnames))
			break;
		name = strtok(linebuf, "\r\n");
		if(name==NULL || *name == '\0')
			continue;

		tmp = SDL_LoadBMP(name);
		if(tmp==NULL) {
			printf(" - Couldn't load object bitmap '%s'!\n", name);
		}
		else {
			obj[opt.numobjtypes++] = Encode32(tmp, true);
			printf(" - Loaded object bitmap '%s'\n", name);
		}
		if(opt.numobjtypes==MAXOBJ) {
			printf("Warning: Object limit of %d reached; no more object bitmaps will be loaded.\n", MAXOBJ);
			break;
		}
	}
    fclose(objnames);

	if(opt.numobjtypes==0) {
		printf("No objects loaded!\n");
		exit(1);
	}
	else printf("%d object bitmap(s) loaded.\n", opt.numobjtypes);

    // Initialize screen parameters
    RefHeight = W_HEIGHT / ASPECT;
    ObjHeight = RefHeight;
    rec.x = 0;
    rec.y = 0;
    rec.w = W_WIDTH;
    rec.h = W_HEIGHT;
}

World::~World()
{
    // delete the maze
    delete[] world;

    // free textures
    for(int i=0; i<NUMTEX; i++)
        SDL_FreeSurface(tex[i]);

}

void World::Resize(int x, int y)
{
	if(world != NULL) delete[] world;
    width = x;
    height = y;
    world = new byte[width * height];
    Clear();
}

// This is the function that actually casts rays.  For efficiency,
// I only check the rays' intercepts with the grid lines.  Even still,
// this function isn't terribly polished, and probably not as efficient
// as it could be; however, since it works, and it accounts for < 5% of 
// execution time, I found no great motivation to revise it.

void World::Render3D(float x, float y, int vangle)
{
    int a, h, t, th, top, bot;
    float x0, dx, y0, dy;
    float dst_x, dst_y, dst;
    float tx_x, tx_y, tx;
    int ix, iy, dix, diy, dc;
    int tex_x, tex_y;
    int col, a1, angle, va;

    for(col=0, a1=vangle+FOV_L; col < W_WIDTH; col++, a1--)
    {
        angle = a1;
        if(angle >= CIRCLE) angle -= CIRCLE;
        else if(angle < D0) angle += CIRCLE;

        va = ABS(a1 - vangle);

        dst_x = INFINITY;
        dst_y = INFINITY;

        if(angle==D0)          // straight right
        {
            iy = (int)y;
            for(ix = (int)x + 1; (tex_y=get(ix, iy))==0; ix++);
            dst_y = ((float)ix - x) * cos_t[va];
            tx_y = FPART(y);
            goto plot;
        }
        else if(angle < D90)   // first quadrant
        {
            a = angle;

            x0 = x + FPART(y) * tan_recip_t[a];
            dx = tan_recip_t[a];
            iy = (int)y - 1;
            diy = -1;

            y0 = y - (1.0f-FPART(x)) * tan_t[a];
            dy = -tan_t[a];
            ix = (int)x + 1;
            dix = 1;
        }
        else if(angle == D90)  // straight up
        {
            ix = (int)x;
            for(iy = (int)y - 1; (tex_x=get(ix, iy))==0; iy--);
            dst_x = (y - (float)(iy+1)) * cos_t[va];
            tx_x = FPART(x);
            goto plot;
        }
        else if(angle < D180)  // second quadrant
        {
            a = D180 - angle;

            x0 = x - FPART(y) * tan_recip_t[a];
            dx = -tan_recip_t[a];
            iy = (int)y - 1;
            diy = -1;

            y0 = y - FPART(x) * tan_t[a];
            dy = -tan_t[a];
            ix = (int)x - 1;
            dix = -1;
        }
        else if(angle == D180) // straight left
        {
            iy = (int)y;
            for(ix = (int)x - 1; (tex_y=get(ix, iy))==0; ix--);
            dst_y = (x - (float)(ix+1)) * cos_t[va];
            tx_y = 1.0f - FPART(y);
            goto plot;
        }
        else if(angle < D270)  // third quadrant
        {
            a = angle - D180;

            x0 = x - (1.0f - FPART(y)) * tan_recip_t[a];
            dx = -tan_recip_t[a];
            iy = 1 + (int)y;
            diy = 1;

            y0 = y + FPART(x) * tan_t[a];
            dy = tan_t[a];
            ix = (int)x - 1;
            dix = -1;
        }
        else if(angle == D270) // straight down
        {
            ix = (int)x;
            for(iy = (int)y + 1; (tex_x=get(ix, iy))==0; iy++);
            dst_x = ((float)iy - y) * cos_t[va];
            tx_x = 1.0f - FPART(x);
            goto plot;
        }
        else                   // fourth quadrant
        {
            a = CIRCLE - angle;

            x0 = x + (1.0f - FPART(y)) * tan_recip_t[a];
            dx = tan_recip_t[a];
            iy = 1 + (int)y;
            diy = 1;

            y0 = y + (1.0f - FPART(x)) * tan_t[a];
            dy = tan_t[a];
            ix = 1 + (int)x;
            dix = 1;
        }

        // trace col-intercepts
        if(((int)x0 < width) && ((int)x0 >= 0)) {
            while((tex_x=get((int)x0, iy))==0) {
                iy += diy;
                x0 += dx;
                if((int)x0 >= width || (int)x0 < 0)
                    goto next;
            }
            dst_x = (ABS(x0 - x) * cos_t[va]) / cos_t[a];
            tx_x = (diy < 0 ) ? FPART(x0) : 1.0f - FPART(x0);
        }

next:
        // trace row-intercepts
        if(((int)y0 < height) && ((int)y0 >= 0)) {
            while((tex_y=get(ix, (int)y0))==0) {
                ix += dix;
                y0 += dy;
                if((int)y0 >= height || (int)y0 < 0)
                    goto plot;
            }
            dst_y = (ABS(y0 - y) * cos_t[va]) / sin_t[a];
            tx_y = (dix > 0) ? FPART(y0) : 1.0f - FPART(y0);
        }


plot:
        // determine which intersection was closer
        if(dst_x < dst_y) {
            th = (int)(RefHeight / dst_x);
            t = tex_x;
            tx = tx_x;
            dst = dst_x;
            Zbuffer[col] = dst_x;
            dc = 0;
        }
        else if(dst_y < INFINITY) {
            th = (int)(RefHeight / dst_y);
            t = tex_y;
            tx = tx_y;
            dst = dst_y;
            Zbuffer[col] = dst_y;
            dc = 40;
        }
    	else return;	// neither ray struck a wall--this won't happen if the maze is closed

        h = (th < W_HEIGHT) ? th : W_HEIGHT;

        // plot the column!!!
        top = W_VMID-(h>>1); bot = top + h;
        sky_gradient(col, top);
        DrawStars(angle, col);
        texturecol(col, top, h, tx, th, dst, dc, tex[t - 1]);
        ground_gradient(col, W_HEIGHT - bot);
    } // for(each column on screen)
}

void World::Clear(int tex)
{
    int i, j;

    for(i=0; i<width; i++)
        for(j=0; j<height; j++)
            set(i, j, tex);
}

void World::CreateTitle()
{
	int i, x, y;

	// initialize the world
    Resize(6, 7);
	Clear(1);

	// draw the little 'A'
	for(i = 1; i < 5; i++)
	{
		set(1, i, 0);
		set(4, i, 0);
		set(i, 1, 0);
		set(i, 3, 0);
	}
	set(1, 5, 0);
	set(4, 5, 0);

    // Initialize player
    x = 1; y = 2; state.angle = D270;
    setobj(x, y);
    state.x = 0.50f + x;
    state.y = 0.50f + y;
    state.v = 0.0f;
    state.vs = 0.0f;
    state.w = 0;
    state.map = 1;
    state.fps = 0;
    state.run = 1.0f;
    state.collected = 0;
    state.time = 0;

    // Create an object to look at
    numobjects = opt.collectables = 1;
    objects[0].Init(0, 1.50f, 3.50f);

    // Place an exit
    set(0, 3, 2);
    opt.exit_x = 1;
    opt.exit_y = 5;

}

void World::CreateMaze(int tex)
{
    int x, y;

#ifdef EMPTYMAZE
    // create an empty box (for testing purposes)
    Clear(0);
    for(x=0; x<width; x++)
    {
        set(x, 0, tex);
        set(x, height-1, tex);
    }
    for(y=1; y<height-1; y++)
    {
        set(0, y, tex);
        set(width-1, y, tex);
    }
#else
    // We'll actually fill the world solid, and then bore out a maze
    Clear(tex);
    BurrowMaze(1, 1);

    // Post-process to remove all one-block "islands"
    for(x=2; x<(width-2); x++)
        for(y=2; y<(height-2); y++)
            if(get(x,y) && !get(x-1, y) && !get(x+1, y) && !get(x, y-1) && !get(x, y+1))
            {
                switch(rand() & 3) {
                case 0: set(x-1, y, tex); break;
                case 1: set(x+1, y, tex); break;
                case 2: set(x, y-1, tex); break;
                case 3: set(x, y+1, tex); break;
                }
            }
#endif

    // Initialize objects
    int i;

    opt.collectables = numobjects = (width-1)/2;
    for(i=0; i<opt.collectables; i++)
    {
       randopen(x, y);
       setobj(x, y);

	   objects[i].Init(i % opt.numobjtypes, 0.50f + x, 0.50f + y);
    }

    // Initialize player
    randopen(x, y);
    setobj(x, y);
    state.x = 0.50f + x;
    state.y = 0.50f + y;
    state.v = 0.0f;
    state.vs = 0.0f;
    state.angle = D90 * (rand() % 4) + 1;
    state.w = 0;
    state.map = 1;
    state.fps = 0;
    state.run = 1.0f;
    state.collected = 0;
    state.time = 0;

	// Put an exit somewheres in this maze
	switch(rand() & 3)
	{
	case 0: // top
		x = 1 + rand() % (width - 1);
		y = 0;
		while(get(x, 1)) {
			x++;
			if(x >= (width-1)) x = 1;
		}
		break;
	case 1: // bottom
		x = 1 + rand() % (width - 1);
		y = height - 1;
		while(get(x, y-1)) {
			x++;
			if(x >= (width-1)) x = 1;
		}
		break;
	case 2: // left
		y = 1 + rand() % (height - 1);
		x = 0;
		while(get(1, y)) {
			y++;
			if(y >= (height-1)) y = 1;
		}
		break;
	case 3: // right
		y = 1 + rand() % (height - 1);
		x = width - 1;
		while(get(x - 1, y)) {
			y++;
			if(y >= (height-1)) y = 1;
		}
		break;
	}
	set(x, y, 2);
    opt.exit_x = x;
    opt.exit_y = y;
}

void World::randopen(int &x, int &y)
{
    x = rand() % width;
    y = rand() % height;

    while(getobj(x, y) != 0)
    {
       x++; if(x >= width) x = 0;
       y++; if(y >= width) y = 0;
    }
}

// essentially a weighted coin BurrowMaze uses to decide whether to make a redundant tunnel
int World::openthis()
{
    return (rand() % OPENNESS) == 0;
}

// Recursive component of random maze generator
void World::BurrowMaze(int x, int y)
{
    // check if we've hit a boundary
    if(x < 1 || y < 1 || x >= (width-1) || y >= (height-1))
        return;

    // check if we've been here before
    if(get(x,y)==0)
        return;

    // burrow out this square
    set(x, y, 0);

    // burrow neighbors in random order
    int pos, nswp, i, j, arr[4] = {0, 1, 2, 3};

    for(nswp=0; nswp < 8; nswp++)
    {
        i = rand() & 3;
        j = rand() & 3;

        pos = arr[i];
        arr[i] = arr[j];
        arr[j] = pos;
    }

    for(pos=0; pos<4; pos++)
    {
        switch(arr[pos])
        {
        case 0:   // burrow right
            if(x < (width-2)) {
                if(get(x+2, y) || openthis()) {
                    set(x+1, y, 0);
                    BurrowMaze(x+2, y);
                }
            }
            break;
        case 1:    // burrow left
            if(x > 1) {
                if(get(x-2, y) || openthis()) {
                    set(x-1, y, 0);
                    BurrowMaze(x-2, y);
                }
            }
            break;
        case 2:    // burrow up
            if(y > 1) {
                if(get(x, y-2) || openthis()) {
                    set(x, y-1, 0);
                    BurrowMaze(x, y-2);
                }
            }
            break;
        case 3:    // burrow down
            if(y < (height-2)) {
                if(get(x, y+2) || openthis()) {
                    set(x, y+1, 0);
                    BurrowMaze(x, y+2);
                }
            }
            break;
        }
    }
}

// Draws a translucent 2D representation of the maze at (u, v), with block width bw
void World::Quick2DTrans(int u, int v, int bw)
{
    int x, y, t;
    int pw = (bw < 6) ? 1 : 2;

    if(opt.skill==3)
        return;         // no map allowed on 'advanced' level

    for(x = 0; x < width; x++) {
        for(y = 0; y < height; y++) {
            t = get(x, y);
            if(t==1 || (t > 1) && (opt.skill > 0)) {
                DrawBox_d(u + bw * x, v + bw * y, bw, bw, TRANSR, TRANSG, TRANSB);
            }
        }
    }

    if(opt.skill==2)
       return;        // no annotaions allowed on 'intermediate' level

    // Draw objects on map
    u--;
    v--;

    int a, b;

    for(x = 0; x < numobjects; x++) {
       if(objects[x].IsVisible()) {
          a = (int)(objects[x].GetX() * bw + 0.5f + u);
          b = (int)(objects[x].GetY() * bw + 0.5f + v);
          DrawBox(a, b, pw, pw, 255, 0, 0);
       }
    }

    if(opt.skill==1)
       return;       // no player position on 'novice' level

    // Draw player on map
    a = (int)(state.x * bw + 0.5f + u);
    b = (int)(state.y * bw + 0.5f + v);
    DrawBox(a, b, pw, pw, 255, 255, 0);

}

// This function prepares objects to be drawn by determining their visibility, computing their screen coordinates
// and sizes, and then passing them to DrawObject (in graphics.cpp) in the correct order (decreasing distance)
void World::RenderObjects()
{
    float x0, y0, x, y;
    int a, h, h2, u, v;
    int vcnt = 0;

    int i, j;
    float delta = opt.mv;
    if(delta < 0.15f) delta = 0.15f;

    for(i=0; i < numobjects; i++)
    {
        // See if the object is visible
        if(!objects[i].IsVisible())
           continue;

        // Retrieve object's coordinates
	x = objects[i].GetX();
	y = objects[i].GetY();

        // Translate coordinates to place player at (0, 0)
        x0 = x - state.x;
        y0 = y - state.y;

        // Rotate coordinates to align with player's view
        a = state.angle;
        x = x0 * cos_t[a] - y0 * sin_t[a];
        y = x0 * sin_t[a] + y0 * cos_t[a];
        if(x <= 0.0f || x > 9.0f) continue;     // the object is out of range

        // Perform viewport transformation
        h = (int)(ObjHeight / x);
        h2 = h >> 1;
        u = W_HMID + (int)((y * W_HMID) / (x * tan_t[FOV_L])) - h2;
        if(u <= -h || u >= W_WIDTH) continue;                          // object is entirely offscreen
        v = W_VMID - h2;

        // since I don't want to clip vertically, I'll just consider objects too large to fit on
        // the screen close enough to be "picked up".  Free collission test!  Ha ha ha ha!!
        if(h > W_HEIGHT)
        {
		state.collected++;
                objects[i].SetVisible(0);
			if(state.collected==opt.collectables)
				set(opt.exit_x, opt.exit_y, 3);
                continue;
        }

        // Reserve object's place in the queue
        queue[vcnt].u = u;
        queue[vcnt].v = v;
        queue[vcnt].d = x;
        queue[vcnt].w = h;
        queue[vcnt].t = objects[i].GetType();
        vcnt++;
    }

    // Now plot the objects from farthest to closest
    // (This will ensure objects are occluded properly by other objects)
    // Yes, it's a selection sort, but the number of visible objects at any one time isn't likely to be terribly large
    // [note: profiling reveals that this entire function, including rotation and translation above,
    //  accounts for less than 1/100 of 1% of total execution time, so there's nothing to be saved by using a
    //  smarter sort algorithm, unless perhaps hundreds of objects are placed in the maze...]
    float maxd;
    int ix;

    for(i=0; i<vcnt; i++)
    {
        maxd = queue[0].d;
        ix = 0;

        // get the farthest object
        for(j=1; j<vcnt; j++)
        {
            if(queue[j].d > maxd) {
                maxd = queue[j].d;
                ix = j;
            }
        }

        // draw it
        DrawObject(queue[ix].u, queue[ix].v, queue[ix].w, queue[ix].d, obj[queue[ix].t]);

        // ensure it won't be chosen again
        queue[ix].d = 0;
    }

}

